# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2021 Osyris contributors (https://github.com/nvaytet/osyris)

# flake8: noqa

from .array import Array
from .dataset import Dataset
from .plot import Plot



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2021 Osyris contributors (https://github.com/nvaytet/osyris)

import numpy as np
from ..config import parameters, additional_variables
from . import utils
from ..core import Dataset
from .amr import AmrLoader
from .grav import GravLoader
from .hydro import HydroLoader
from .rt import RtLoader
from .units import get_unit


def load(nout=1, scale=None, path="", select=None, cpu_list=None, bounding_box=None):

    data = Dataset()

    if select is None:
        select = {}

    if scale is None:
        scale = parameters["scale"]

    # Generate directory name from output number
    infile = utils.generate_fname(nout, path)

    # Read info file and create info dictionary
    infofile = infile + "/info_" + infile.split("_")[-1] + ".txt"
    data.meta.update(utils.read_parameter_file(fname=infofile))

    # Add additional information
    data.meta["scale"] = scale
    data.meta["infile"] = infile
    data.meta["path"] = path
    data.meta["time"] = data.meta["time"] * get_unit(
        "time", data.meta["unit_d"], data.meta["unit_l"], data.meta["unit_t"])

    # Take into account user specified lmax
    if "level" in select:
        data.meta["lmax"] = utils.find_max_amr_level(levelmax=data.meta["levelmax"],
                                                     select=select)
    else:
        data.meta["lmax"] = data.meta["levelmax"]

    code_units = {
        "ud": data.meta["unit_d"],
        "ul": data.meta["unit_l"],
        "ut": data.meta["unit_t"]
    }

    loader_list = {
        "amr":
        AmrLoader(scale=scale,
                  select=select,
                  code_units=code_units,
                  meta=data.meta,
                  infofile=infofile),
        "hydro":
        HydroLoader(infile=infile, select=select, code_units=code_units),
        "grav":
        GravLoader(nout=nout,
                   path=path,
                   select=select,
                   units=code_units,
                   ndim=data.meta["ndim"]),
        "rt":
        RtLoader(infile=infile, select=select, units=code_units)
    }

    # Initialize loaders
    loaders = {}
    for group, loader in loader_list.items():
        if loader.initialized:
            loaders[group] = loader

    # Take into account user specified cpu list
    if cpu_list is None:
        cpu_list = loader_list["amr"].cpu_list if loader_list[
            "amr"].cpu_list is not None else range(1, data.meta["ncpu"] + 1)

    print("Processing {} files in {}".format(len(cpu_list), infile))

    # Allocate work arrays
    twotondim = 2**data.meta["ndim"]
    for loader in loaders.values():
        loader.allocate_buffers(ngridmax=data.meta["ngridmax"], twotondim=twotondim)

    iprog = 1
    istep = 10
    ncells_tot = 0
    npieces = 0

    # integer, double, line, string, quad, long
    null_offsets = {key: 0 for key in "idnsql"}

    # Loop over the cpus and read the AMR and HYDRO files in binary format
    for cpu_ind, cpu_num in enumerate(cpu_list):

        # Print progress
        percentage = int(float(cpu_ind) * 100.0 / float(len(cpu_list)))
        if percentage >= iprog * istep:
            print("{:>3d}% : read {:>10d} cells".format(percentage, ncells_tot))
            iprog += 1

        # Read binary files
        for group, loader in loaders.items():
            fname = utils.generate_fname(nout, path, ftype=group, cpuid=cpu_num)
            with open(fname, mode='rb') as f:
                loader.bytes = f.read()

        # Read file headers
        for loader in loaders.values():
            loader.offsets.update(null_offsets)
            loader.read_header(data.meta)

        # Loop over levels
        for ilevel in range(data.meta["lmax"]):

            for loader in loaders.values():
                loader.read_level_header(ilevel, twotondim)

            # Loop over domains
            for domain in range(loaders["amr"].meta["nboundary"] + data.meta["ncpu"]):

                ncache = loaders["amr"].meta["ngridlevel"][domain, ilevel]

                for loader in loaders.values():
                    loader.read_domain_header()

                if ncache > 0:

                    if domain == cpu_num - 1:

                        for loader in loaders.values():
                            loader.read_cacheline_header(ncache, data.meta["ndim"])

                        for ind in range(twotondim):

                            # Read variables in cells
                            for loader in loaders.values():
                                loader.read_variables(ncache, ind, ilevel, cpu_num - 1,
                                                      data.meta)

                        # Apply selection criteria: select only leaf cells and
                        # add any criteria requested by the user via select.
                        conditions = {}
                        for loader in loaders.values():
                            conditions.update(loader.make_conditions(select, ncache))
                        # Combine all selection criteria together with AND
                        # operation by using a product on bools
                        sel = np.where(
                            np.prod(np.array(list(conditions.values())), axis=0))

                        # Count the number of cells
                        ncells = np.shape(sel)[1]
                        if ncells > 0:
                            ncells_tot += ncells
                            npieces += 1
                            # Add the cells in the pieces dictionaries
                            for loader in loaders.values():
                                for item in loader.variables.values():
                                    item["pieces"][npieces] = item["buffer"][sel]

                        # Increment offsets with remainder of the file
                        for loader in loaders.values():
                            loader.read_footer(ncache, twotondim)

                    else:

                        for loader in loaders.values():
                            loader.step_over(ncache, twotondim, data.meta["ndim"])

    # Store the number of cells
    data.meta["ncells"] = ncells_tot

    # Merge all the data pieces into the Arrays
    for group in loaders.values():
        for key, item in group.variables.items():
            data[key] = np.concatenate(list(item["pieces"].values()))

    # If vector quantities are found, make them into vector Arrays
    utils.make_vector_arrays(data)

    # Create additional variables derived from the ones already loaded
    additional_variables(data)

    print("Total number of cells loaded: {}".format(ncells_tot))
    print("Memory used: {}".format(data.print_size()))

    return data



import numpy as np
from pint.quantity import Quantity
from pint.unit import Unit
from .tools import value_to_string, make_label
from .. import units


def _comparison_operator(lhs, rhs, op):
    func = getattr(np, op)
    if isinstance(rhs, Array):
        scale_r = rhs.unit.to(lhs._unit.units)
        return func(lhs._array, rhs._array * scale_r.magnitude)
    if isinstance(rhs, Quantity):
        return func(lhs._array, rhs.to(lhs._unit.units).magnitude)
    return func(lhs._array, rhs)


class Array:
    def __init__(self, values=None, unit=None, parent=None, name=""):

        if isinstance(values, np.ndarray):
            self._array = values
        else:
            self._array = np.asarray(values)

        if unit is None:
            self._unit = 1.0 * units.dimensionless
        elif isinstance(unit, str):
            self._unit = units(unit)
        elif isinstance(unit, Quantity):
            self._unit = unit
        elif isinstance(unit, Unit):
            self._unit = 1.0 * unit
        else:
            raise TypeError("Unsupported unit type {}".format(type(unit)))
        self._parent = parent
        self._name = name
        if self._array.shape:
            self._ndim = self._array.shape[-1]
        else:
            self._ndim = 0
        self.special_functions = ["sqrt", "power"]

    # def __array__(self):
    #     return self._array

    def __getitem__(self, slice_):
        return self.__class__(values=self._array[slice_],
                              unit=self._unit,
                              parent=self._parent,
                              name=self._name)

    def __len__(self):
        if self._array.shape:
            return len(self._array)
        else:
            return 0

    def __str__(self):
        name_str = "'" + self._name + "' "
        if len(self) == 0:
            values_str = "Value: " + value_to_string(self.values)
        else:
            values_str = "Min: " + value_to_string(
                self.values.min()) + " Max: " + value_to_string(self.values.max())
        unit_str = " [{:~}] ".format(self._unit.units)
        shape_str = str(self._array.shape)
        return name_str + values_str + unit_str + shape_str

    def __repr__(self):
        return str(self)

    def copy(self):
        return self.__class__(values=self._array.copy(),
                              unit=self._unit.copy(),
                              name=self._name)

    @property
    def values(self):
        return self._array

    @values.setter
    def values(self, values_):
        self._array = values_

    @property
    def array(self):
        return self._array

    @array.setter
    def array(self, array_):
        self._array = array_

    @property
    def norm(self):
        if self._array.ndim < 2:
            return self._array
        else:
            return np.linalg.norm(self._array, axis=1)

    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, unit_):
        self._unit = unit_

    @property
    def ndim(self):
        return self._ndim

    @property
    def shape(self):
        return self._array.shape

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, parent_):
        self._parent = parent_

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name_):
        self._name = name_

    @property
    def x(self):
        if self.ndim > 1:
            return self.__class__(values=self._array[:, 0],
                                  unit=self._unit,
                                  parent=self._parent,
                                  name=self._name + "_x")

    @property
    def y(self):
        if self.ndim > 1:
            return self.__class__(values=self._array[:, 1],
                                  unit=self._unit,
                                  parent=self._parent,
                                  name=self._name + "_y")

    @property
    def z(self):
        if self.ndim > 2:
            return self.__class__(values=self._array[:, 2],
                                  unit=self._unit,
                                  parent=self._parent,
                                  name=self._name + "_z")

    @property
    def label(self):
        return make_label(name=self._name, unit=self._unit.units)

    def _broadcast(self, lhs, rhs):
        if lhs.ndim == rhs.ndim:
            return lhs, rhs
        if lhs.ndim > rhs.ndim:
            ind = np.argmax(np.array(lhs.shape) == rhs.shape[0])
            if ind == 0:
                return lhs, rhs.reshape(rhs.shape + tuple([1]))
            else:
                return lhs, rhs.reshape(tuple([1]) + rhs.shape)
        else:
            ind = np.argmax(np.array(rhs.shape) == lhs.shape[0])
            if ind == 0:
                return lhs.reshape(lhs.shape + tuple([1])), rhs
            else:
                return lhs.reshape(tuple([1]) + lhs.shape), rhs

    def _raise_incompatible_units_error(self, other, op):
        raise TypeError("Could not {} types {} and {}.".format(op, self, other))

    def __add__(self, other):
        if isinstance(other, self.__class__):
            scale_r = other.unit.to(self._unit.units)
            lhs = self._array
            rhs = other._array * scale_r.magnitude
            lhs, rhs = self._broadcast(lhs, rhs)
            return self.__class__(values=lhs + rhs, unit=self._unit)
        if isinstance(other, Quantity):
            return self.__class__(values=self._array +
                                  other.to(self._unit.units).magnitude,
                                  unit=self._unit)
        self._raise_incompatible_units_error(other, "add")

    def __iadd__(self, other):
        if isinstance(other, self.__class__):
            scale_r = other.unit.to(self._unit.units)
            rhs = other._array * scale_r.magnitude
            self._array += rhs
        elif isinstance(other, Quantity):
            self._array += other.to(self._unit.units).magnitude
        else:
            self._raise_incompatible_units_error(other, "add")
        return self

    def __sub__(self, other):
        if isinstance(other, self.__class__):
            scale_r = other.unit.to(self._unit.units)
            lhs = self._array
            rhs = other._array * scale_r.magnitude
            lhs, rhs = self._broadcast(lhs, rhs)
            return self.__class__(values=lhs - rhs, unit=self._unit)
        if isinstance(other, Quantity):
            return self.__class__(values=self._array -
                                  other.to(self._unit.units).magnitude,
                                  unit=self._unit)
        self._raise_incompatible_units_error(other, "subtract")

    def __isub__(self, other):
        if isinstance(other, self.__class__):
            scale_r = other.unit.to(self._unit.units)
            rhs = other._array * scale_r.magnitude
            self._array -= rhs
        elif isinstance(other, Quantity):
            self._array -= other.to(self._unit.units).magnitude
        else:
            self._raise_incompatible_units_error(other, "subtract")
        return self

    def __mul__(self, other):
        if isinstance(other, self.__class__):
            scale_l = self._unit.to_base_units()
            scale_r = other._unit.to_base_units()
            result = scale_l * scale_r
            lhs = self._array
            rhs = other._array * result.magnitude
            lhs, rhs = self._broadcast(lhs, rhs)
            return self.__class__(values=lhs * rhs, unit=1.0 * result.units)
        if isinstance(other, Quantity):
            scale_l = self._unit.to_base_units()
            scale_r = other.to_base_units()
            result = scale_l * scale_r
            return self.__class__(values=self._array * result.magnitude,
                                  unit=1.0 * result.units)
        return self.__class__(values=self._array * other, unit=self._unit)

    def __imul__(self, other):
        if isinstance(other, self.__class__):
            scale_l = self._unit.to_base_units()
            scale_r = other._unit.to_base_units()
            result = scale_l * scale_r
            rhs = other._array * result.magnitude
            self._array *= rhs
            self._unit = 1.0 * result.units
        elif isinstance(other, Quantity):
            scale_l = self._unit.to_base_units()
            scale_r = other.to_base_units()
            result = scale_l * scale_r
            self._array *= result.magnitude
            self._unit = 1.0 * result.units
        else:
            self._array *= other
        return self

    def __truediv__(self, other):
        if isinstance(other, self.__class__):
            scale_l = self._unit.to_base_units()
            scale_r = other._unit.to_base_units()
            result = scale_l / scale_r
            lhs = self._array
            rhs = other._array / result.magnitude
            lhs, rhs = self._broadcast(lhs, rhs)
            return self.__class__(values=lhs / rhs, unit=1.0 * result.units)
        if isinstance(other, Quantity):
            scale_l = self._unit.to_base_units()
            scale_r = other.to_base_units()
            result = scale_l / scale_r
            return self.__class__(values=self._array * result.magnitude,
                                  unit=1.0 * result.units)
        return self.__class__(values=self._array / other, unit=self._unit)

    def __itruediv__(self, other):
        if isinstance(other, self.__class__):
            scale_l = self._unit.to_base_units()
            scale_r = other._unit.to_base_units()
            result = scale_l / scale_r
            rhs = other._array / result.magnitude
            self._array /= rhs
            self._unit = 1.0 * result.units
        elif isinstance(other, Quantity):
            scale_l = self._unit.to_base_units()
            scale_r = other.to_base_units()
            result = scale_l / scale_r
            self._array *= result.magnitude
            self._unit = 1.0 * result.units
        else:
            self._array /= other
        return self

    def __rmul__(self, other):
        return self * other

    def __rtruediv__(self, other):
        if isinstance(other, self.__class__):
            scale_r = self._unit.to_base_units()
            scale_l = other._unit.to_base_units()
            result = scale_l / scale_r
            lhs = self._array
            rhs = other._array / result.magnitude
            lhs, rhs = self._broadcast(lhs, rhs)
            return self.__class__(values=lhs / rhs, unit=1.0 * result.units)
        if isinstance(other, Quantity):
            scale_r = self._unit.to_base_units()
            scale_l = other.to_base_units()
            result = scale_l / scale_r
            return self.__class__(values=self._array * result.magnitude,
                                  unit=1.0 * result.units)
        return self.__class__(values=other / self._array, unit=1.0 / self._unit)

    def __pow__(self, number):
        return np.power(self, number)

    def __lt__(self, other):
        return _comparison_operator(self, other, "less")

    def __le__(self, other):
        return _comparison_operator(self, other, "less_equal")

    def __gt__(self, other):
        return _comparison_operator(self, other, "greater")

    def __ge__(self, other):
        return _comparison_operator(self, other, "greater_equal")

    def __eq__(self, other):
        return _comparison_operator(self, other, "equal")

    def __ne__(self, other):
        return _comparison_operator(self, other, "not_equal")

    def to(self, unit):
        new_unit = units(unit)
        ratio = self._unit.to(new_unit) / new_unit
        self._unit = 1.0 * new_unit
        self._array *= ratio.magnitude
        return self

    def _wrap_numpy(self, func, *args, **kwargs):
        if func.__name__ in self.special_functions:
            unit = func(self.unit, *args[1:], **kwargs)
        else:
            unit = self.unit
        if isinstance(args[0], tuple) or isinstance(args[0], list):
            # Case where we have a sequence of arrays, e.g. `concatenate`
            for a in args[0]:
                if a.unit != unit:
                    self._raise_incompatible_units_error(a, func.__name__)
            args = (tuple(a._array for a in args[0]), ) + args[1:]
        elif (len(args) > 1 and hasattr(args[1], "_array")):
            if hasattr(args[0], "_array"):
                # Case of a binary operation, with two Arrays, e.g. `dot`
                # TODO: what should we do with the unit? Apply the func to it?
                args = (args[0]._array, args[1]._array) + args[2:]
                unit = func(args[0].unit, args[1].unit, *args[2:], **kwargs)
            else:
                # Case of a binary operation: ndarray with Array
                # In this case, only multiply is allowed?
                if func.__name__ != "multiply":
                    raise RuntimeError("Cannot use operation {} between ndarray and "
                                       "Array".format(func.__name__))
                args = (args[0], args[1]._array) + args[2:]
        else:
            args = (args[0]._array, ) + args[1:]
        result = func(*args, **kwargs)
        return self.__class__(values=result, unit=unit)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """
        Numpy array_ufunc protocol to allow Array to work with numpy ufuncs.
        """
        if method != "__call__":
            # Only handle ufuncs as callables
            return NotImplemented
        return self._wrap_numpy(ufunc, *inputs, **kwargs)

    def __array_function__(self, func, types, args, kwargs):
        """
        Numpy array_function protocol to allow Array to work with numpy
        functions.
        """
        return self._wrap_numpy(func, *args, **kwargs)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2021 Osyris contributors (https://github.com/nvaytet/osyris)

# flake8: noqa

from .histogram import histogram
from .render import render
from .plane import plane



import numpy as np
from ..config import parameters


def quiver(ax, x, y, z, density=1, color="w", **kwargs):
    """
    Wrapper around Matplotlib's quiver plot.

    We have a special argument `density` which aims to control the density of
    arrows in a similar manner to the density of streamlines in the streamplot.
    Matplotlib's own way of controlling the density of arrows is not so
    obvious.
    """
    default_args = {
        "angles": "xy",
        "pivot": "mid",
    }
    default_args.update(kwargs)

    skips = np.around(np.array(z.shape) * 4.0 / 128.0 / density).astype(np.int)
    skip = (slice(None, None, skips[0]), slice(None, None, skips[1]))

    args = [x[skip[0]], y[skip[1]], z[..., 0][skip], z[..., 1][skip]]
    if isinstance(color, str):
        default_args["color"] = color
    else:
        args.append(z[..., 2][skip])

    return ax.quiver(*args, **default_args)


def pcolormesh(ax, x, y, z, **kwargs):
    """
    Wrapper around Matplotlib's pcolormesh plot.
    """
    default_args = {
        "shading": "nearest",
    }
    default_args.update(kwargs)
    if "cmap" not in kwargs:
        kwargs["cmap"] = parameters["cmap"]
    return ax.pcolormesh(x, y, z, **default_args)


def contour(ax, x, y, z, labels=True, **kwargs):
    """
    Wrapper around Matplotlib's contour plot.

    We add a small convenience argument `labels` that allows to simply add
    clabels with a simple `labels=True`.
    """
    cs = ax.contour(x, y, z, **kwargs)
    if labels:
        ax.clabel(cs, inline=1, fontsize=10)
    return cs


def contourf(ax, x, y, z, **kwargs):
    """
    Wrapper around Matplotlib's contourf plot.
    """
    if "cmap" not in kwargs:
        kwargs["cmap"] = parameters["cmap"]
    return ax.contourf(x, y, z, **kwargs)


def streamplot(ax, x, y, z, **kwargs):
    """
    Wrapper around Matplotlib's streamplot plot.
    """
    default_args = {"color": "w"}
    default_args.update(kwargs)
    return ax.streamplot(x, y, z[..., 0], z[..., 1], **default_args)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2021 Osyris contributors (https://github.com/nvaytet/osyris)

# flake8: noqa

# Import the config from "/home/user/.osyris/config if it exists.
# If it doesn't, try to create one by copying the default from the source.
# If that fails, just load the default.
import os
import sys
from pint import UnitRegistry

units = UnitRegistry(system="cgs")

config_dir = os.path.join(os.path.expanduser("~"), ".osyris")
sys.path.append(config_dir)
try:
    import config_osyris as config
except ImportError:
    from shutil import copyfile
    this_dir = os.path.dirname(os.path.abspath(__file__))
    if not os.path.exists(config_dir):
        os.mkdir(config_dir)
    copyfile(os.path.join(this_dir, "config.py"),
             os.path.join(config_dir, "config_osyris.py"))
    try:
        import config_osyris as config
    except ImportError:
        from . import config

config.additional_units(units)

from .io import load
from .plot import histogram, plane
from .core import Array, Dataset, Plot



import numpy as np
from .loader import Loader
from .units import get_unit
from . import utils


class HydroLoader(Loader):
    def __init__(self, infile, select, code_units):

        super().__init__()

        # Read the number of variables from the hydro_file_descriptor.txt
        # and select the ones to be read if specified by user
        self.initialized = True
        fname = infile + "/hydro_file_descriptor.txt"
        try:
            descriptor = np.loadtxt(fname, dtype=str, delimiter=",")
        except IOError:
            self.initialized = False

        if self.initialized:
            for i in range(len(descriptor)):
                key = descriptor[i, 1].strip()
                read = True
                if key in select:
                    if select[key] is False:
                        read = False
                if "hydro" in select:
                    if select["hydro"] is False:
                        read = False
                self.variables[key] = {
                    "read":
                    read,
                    "type":
                    descriptor[i, 2].strip(),
                    "buffer":
                    None,
                    "pieces": {},
                    "unit":
                    get_unit(key, code_units["ud"], code_units["ul"], code_units["ut"])
                }

    def read_header(self, info):
        # hydro gamma
        self.offsets["i"] += 5
        self.offsets["n"] += 5
        [info["gamma"]] = utils.read_binary_data(fmt="d",
                                                 content=self.bytes,
                                                 offsets=self.offsets)

    def read_domain_header(self):
        self.offsets['n'] += 2
        self.offsets['i'] += 2



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2021 Osyris contributors (https://github.com/nvaytet/osyris)

import glob
import struct
import numpy as np
from ..core import Array


def generate_fname(nout, path="", ftype="", cpuid=1, ext=""):

    if len(path) > 0:
        if path[-1] != "/":
            path = path + "/"

    if nout == -1:
        filelist = sorted(glob.glob(path + "output*"))
        number = filelist[-1].split("_")[-1]
    else:
        number = str(nout).zfill(5)

    infile = path + "output_" + number
    if len(ftype) > 0:
        infile += "/" + ftype + "_" + number
        if cpuid >= 0:
            infile += ".out" + str(cpuid).zfill(5)

    if len(ext) > 0:
        infile += ext

    return infile


def read_parameter_file(fname=None, delimiter="="):
    """
    Read info file and create dictionary
    """
    out = {}
    with open(fname) as f:
        content = f.readlines()
    for line in content:
        sp = line.split(delimiter)
        if len(sp) > 1:
            value = sp[1].strip()
            try:
                value = eval(value)
            except NameError:
                pass
            out[sp[0].strip()] = value
    return out


def read_binary_data(content=None,
                     fmt=None,
                     offsets=None,
                     skip_head=True,
                     increment=True):
    """
    Unpack binary data from a content buffer using a dict of offsets.
    Also increment the offsets of the corresponding data read, as well as
    increase the line count by 1.
    """

    byte_size = {
        "b": 1,
        "h": 2,
        "i": 4,
        "q": 8,
        "f": 4,
        "d": 8,
        "e": 8,
        "n": 8,
        "l": 8,
        "s": 1
    }

    offset = 0
    for key in offsets:
        offset += offsets[key] * byte_size[key]
    # if offset is None:
    #     offset = 4*ninteg + 8*(nlines+nfloat+nlongi) + nstrin + nquadr*16
    if skip_head:
        offset += 4  # + correction

    # byte_size = {"b": 1 , "h": 2, "i": 4, "q": 8, "f": 4, "d": 8, "e": 8}
    if len(fmt) == 1:
        mult = 1
    else:
        mult = int(fmt[:-1])
    pack_size = mult * byte_size[fmt[-1]]

    if increment:
        offsets[fmt[-1]] += mult
    offsets["n"] += 1

    return struct.unpack(fmt, content[offset:offset + pack_size])


def make_vector_arrays(data):
    """
    Merge vector components in 2d arrays.
    """
    components = list("xyz"[:data.meta["ndim"]])
    if len(components) > 1:
        skip = []
        for key in list(data.keys()):
            # Make list of 3 components
            comp_list = None
            rawkey = None
            if key.endswith("_x") and key not in skip:
                rawkey = key[:-2]
                comp_list = [rawkey + "_" + c for c in components]
            if "_x_" in key and key not in skip:
                rawkey = key.replace("_x", "")
                comp_list = [key.replace("_x_", "_{}_".format(c)) for c in components]

            if comp_list is not None:
                if all([item in data for item in comp_list]):
                    data[rawkey] = Array(values=np.array(
                        [data[c].values for c in comp_list]).T,
                                         unit=data[key].unit)
                    for c in comp_list:
                        del data[c]
                        skip.append(c)


def find_max_amr_level(levelmax, select):
    """
    Test the selection function in `select` on the range of possible AMR levels
    to determine the max level to read.
    """
    possible_levels = np.arange(1, levelmax + 1, dtype=int)
    func_test = select["level"](possible_levels)
    inds = np.argwhere(func_test).ravel()
    return possible_levels[inds.max()]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2021 Osyris contributors (https://github.com/nvaytet/osyris)
"""
Define default values so that you don't have to specify them every time.
"""
parameters = {
    "scale": "au",
    "path": None,
    "select": None,
    "cmap": "viridis",
    "render_mode": "pcolormesh"
}


def additional_units(ureg):
    """
    Define additional useful units and constants
    """
    ureg.define('solar_mass = 1.9889e+33 * g = msun')
    ureg.define('radiation_constant = 7.56591469318689378e-015 * erg / cm^3 / K^4 = ar')


def additional_variables(data):
    """
    Here are some additional variables that are to be computed every time data
    is loaded.

    It is recommended to place your variables in a `try/except` block, which
    will prevent errors if the variables are not found, for instance when
    loading data from a different simulation.
    """

    # Magnetic field
    try:
        data["B_field"] = 0.5 * (data["B_left"] + data["B_right"])
    except KeyError:
        pass

    # Mass
    try:
        data["mass"] = data["density"] * data["dx"]**3
        data["mass"].to("msun")
    except KeyError:
        pass



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2021 Osyris contributors (https://github.com/nvaytet/osyris)

import matplotlib.pyplot as plt
from ..config import parameters
from . import wrappers
from ..core.tools import make_label


def render(x, y, data, logx=False, logy=False, ax=None):
    """
    Use matplotlib to plot histogram, slice or column density maps
    """

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()

    function_map = {
        "vec": "quiver",
        "vector": "quiver",
        "stream": "streamplot",
        None: parameters["render_mode"],
        "image": "pcolormesh",
        "imshow": "pcolormesh"
    }

    mpl_objects = []
    cbar = None
    for item in data:
        func = item["mode"]
        if func in function_map:
            func = function_map[func]

        mpl_objects.append(
            getattr(wrappers, func)(ax, x, y, item["data"], **item["params"]))

        if func in ["contourf", "pcolormesh"] and cbar is None:
            cbar = plt.colorbar(mpl_objects[-1], ax=ax, cax=None)
            cbar.set_label(make_label(name=item["name"], unit=item["unit"]))
            cbar.ax.yaxis.set_label_coords(-1.1, 0.5)

    if logx:
        ax.set_xscale("log")
    if logy:
        ax.set_yscale("log")

    return {"fig": fig, "ax": ax, "objects": mpl_objects}



import os
from .loader import Loader
from .units import get_unit
from . import utils


class GravLoader(Loader):
    def __init__(self, nout, path, select, units, ndim):

        super().__init__()

        # Check if self-gravity files exist
        fname = utils.generate_fname(nout, path, ftype="grav", cpuid=1)
        self.initialized = os.path.exists(fname)
        # Add gravity fields
        if self.initialized:
            descriptor = {"grav_potential": "d"}
            for n in range(ndim):
                descriptor["grav_acceleration_" + "xyz"[n]] = "d"
            # Now add to the list of variables to be read
            for key in descriptor:
                read = True
                if key in select:
                    if select[key] is False:
                        read = False
                if "gravity" in select:
                    if select["gravity"] is False:
                        read = False
                if "grav" in select:
                    if select["grav"] is False:
                        read = False
                self.variables[key] = {
                    "read": read,
                    "type": descriptor[key],
                    "buffer": None,
                    "pieces": {},
                    "unit": get_unit(key, units["ud"], units["ul"], units["ut"])
                }

    def read_header(self, info):
        self.offsets["i"] += 4
        self.offsets["n"] += 4

    def read_domain_header(self):
        self.offsets['n'] += 2
        self.offsets['i'] += 2
